#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsWorker.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.05.2021
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import gc, signal
from math import ceil
from time import time
from typing import ClassVar, Any, TypeVar, Generic, Literal

from pymongo.write_concern import WriteConcern

from ampel.types import OneOrMany, JDict, UBson, Tag
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.mongo.utils import maybe_match_array
from ampel.log.utils import convert_dollars
from ampel.enum.DocumentCode import DocumentCode
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.enum.EventCode import EventCode
from ampel.log import AmpelLogger, LogFlag, VERBOSE, DEBUG
from ampel.core.EventHandler import EventHandler
from ampel.log.utils import report_exception, report_error
from ampel.log.handlers.DefaultRecordBufferingHandler import DefaultRecordBufferingHandler
from ampel.util.hash import build_unsafe_dict_id
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsUnitResultAdapter import AbsUnitResultAdapter
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.mongo.utils import maybe_use_each
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry, Histogram, Counter
from ampel.util.tag import merge_tags

T = TypeVar("T", T1Document, T2Document)


class AbsWorker(Generic[T], AbsEventUnit, abstract=True):
	"""
	Fill results into pending :class:`T2 documents <ampel.content.T2Document.T2Document>`
	or :class:`T1 documents <ampel.content.T1Document.T1Document>`
	"""

	#: Ids of the units to run. If not specified, any t1/t2 unit will be run.
	unit_ids: None | list[str]

	#: Process only with the given code
	code_match: OneOrMany[int] = [DocumentCode.NEW, DocumentCode.RERUN_REQUESTED]

	#: Max number of docs to process in :func:`run`
	doc_limit: None | int

	#: Tag(s) to add to the stock :class:`~ampel.content.JournalRecord.JournalRecord`
	#: every time a document is processed
	jtag: None | OneOrMany[Tag]

	#: Tag(s) to add to the stock :class:`~ampel.content.MetaRecord.MetaRecord`
	#: every time a document is processed
	mtag: None | OneOrMany[Tag]

	#: Create a 'beacon' document indicating the last time T2Processor was run
	#: in the current configuration
	send_beacon: bool = True

	#: Explicitly call :func:`gc.collect` after the processing of each t1/t2 document.
	#: Turn this on if units do not clean up resources adequately,
	#: but beware of the performance impact.
	garbage_collect: bool = False

	#: Maximum number of processing attempts per document
	max_try: int = 5

	tier: ClassVar[Literal[1, 2]]

	#: Minimum age of documents to be matched (in seconds)
	#: {'$expr': {'$lt': [{'$last': '$meta.ts'}, now - min_doc_age]}}
	min_doc_age: None | float

	#: One day, we might support different DBs
	# database: str = "mongo"

	#: Avoid 'burning' a run_id for nothing at the cost of a request
	pre_check: bool = True

	#: wait for majority acknowledgement of document updates. this introduces
	#: significant extra latency for replicated mongo clusters.
	wait_for_durable_write: bool = True

	#: minimum number of stock document updates to commit at once
	updates_buffer_size: int = 500

	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self._ampel_db = self.context.db
		self._loader = self.context.loader

		# Prepare DB query dict
		self.query: JDict = {
			'code': self.code_match if isinstance(self.code_match, DocumentCode)
				else maybe_match_array(self.code_match) # type: ignore[arg-type]
		}

		# Possibly restrict query to specified t2 units
		if self.unit_ids:
			self.query['unit'] = maybe_match_array(self.unit_ids)

		if self.min_doc_age:
			self.query['$expr'] = {
				'$lt': [{'$last': '$meta.ts'}, time() - self.min_doc_age]
			}

		# Shortcut
		self.col_stock = self._ampel_db.get_collection('stock', mode='r')
		self.col_t0 = self._ampel_db.get_collection('t0', mode='r')
		self.col_t1 = self._ampel_db.get_collection('t1', mode='r')
		self.col = self._ampel_db.get_collection(f't{self.tier}', mode='w')
		if not self.wait_for_durable_write:
			# Only wait for the primary to acknowledge (in-memory) write, rather
			# than a majority of replica set members to acknowledge write to
			# their journals. This has much lower latency, but can result in doc
			# updates being lost if the primary goes down before the write can
			# be replicated, but if this happens the doc can simply be rerun.
			self.col = self.col.with_options(write_concern=WriteConcern(w=1, j=False))

		if self.send_beacon:
			self.create_beacon()

		self._run = True
		self._doc_counter = 0
		signal.signal(signal.SIGTERM, self.sig_exit)
		signal.signal(signal.SIGINT, self.sig_exit)

		# _instances stores unit instances so that they can be re-used in run()
		# Key: set(unit name + config), value: unit instance
		self._instances: JDict = {}

		self._adapters: dict[str, AbsUnitResultAdapter] = {}



	@abstractmethod
	def process_doc(self,
		doc: T, stock_updr: MongoStockUpdater, logger: AmpelLogger
	) -> Any:
		...


	def prepare(self, event_hdlr: EventHandler) -> None | EventCode:
		""" :returns: number of t2 docs processed """
		event_hdlr.set_tier(2)
		if self.pre_check and self.col.count_documents(self.query) == 0:
			return EventCode.PRE_CHECK_EXIT
		return None


	def proceed(self, event_hdlr: EventHandler) -> int:
		""" :returns: number of t2 docs processed """

		event_hdlr.set_tier(self.tier)
		run_id = event_hdlr.get_run_id()

		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, run_id,
			base_flag = getattr(LogFlag, f'T{self.tier}') | LogFlag.CORE | self.base_log_flag
		)

		if self.send_beacon:
			self.col_beacon.update_one(
				{'_id': self.beacon_id},
				{'$set': {'timestamp': int(time())}}
			)

		stock_updr = MongoStockUpdater(
			ampel_db = self.context.db, tier = self.tier, run_id = run_id,
			process_name = self.process_name, logger = logger,
			raise_exc = self.raise_exc, extra_tag = self.jtag
		)

		# Loop variables
		self._doc_counter = 0
		update = {'$set': {'code': DocumentCode.RUNNING}}
		garbage_collect = self.garbage_collect
		doc_limit = self.doc_limit

		# Process docs until next() returns None (breaks condition below)
		self._run = True
		while self._run:

			# get t1/t2 document (code is usually NEW or NEW_PRIO), excluding
			# docs with retry times in the future
			doc = self.col.find_one_and_update(
				self.query | {'$expr': {'$not': {'$lte': [ceil(time()), {'$last': '$meta.retry_after'}]}}},
				update
			)

			# No match
			if doc is None:
				break
			elif logger.verbose > 1:
				logger.debug(f'T{self.tier} doc to process', extra={'doc': doc})

			self.process_doc(doc, stock_updr, logger)

			# Check possibly defined doc_limit
			if doc_limit and self._doc_counter >= doc_limit:
				break

			if garbage_collect:
				gc.collect()

		stock_updr.flush()
		event_hdlr.add_extra(docs=self._doc_counter)

		logger.flush()
		self._instances.clear()

		return self._doc_counter


	def _processing_error(self,
		logger: AmpelLogger, doc: T, body: UBson,
		meta: MetaRecord, msg: None | str = None,
		extra: None | JDict = None, exception: None | Exception = None
	) -> None:
		"""
		- Updates the t1/t2 document by appending a meta entry and
		updating 'code' with value DocumentCode.EXCEPTION if an exception
		was provided through the parameter 'exception' or with value
		DocumentCode.ERROR otherwise.

		- Updates the stock document by appending a JournalRecord to it
		and by updating the 'updated' timestamp

		- Creates a 'trouble' document in the troubles collection to report the incident

		:param msg: added to meta entry if provided
		"""

		if (
			not meta['activity'] or
			(len(meta['activity']) == 1 and meta['activity'][0]['action'] == 0)
		):
			meta['activity'] = []

		if not [el for el in meta['activity'] if (el['action'] & MetaActionCode.SET_CODE)]:
			meta['activity'].append( # type: ignore[attr-defined]
				{'action': MetaActionCode.SET_CODE}
			)

		if 'extra' not in meta:
			meta['extra'] = {}

		meta['extra']['msg'] = msg

		self.commit_update({'_id': doc['_id']}, meta, logger, code=DocumentCode.EXCEPTION) # type: ignore[typeddict-item]

		info: JDict = (extra or {}) | meta | {'stock': doc['stock'], 'doc': doc}
		if exception:
			report_exception(self._ampel_db, logger=logger, exc=exception, info=info)
		else:
			report_error(self._ampel_db, logger=logger, msg=msg, info=info)


	def commit_update(self,
		match: JDict,
		meta: MetaRecord,
		logger: AmpelLogger, *,
		payload_op: Literal['$push', '$set'] = '$push',
		body: UBson = None,
		tag: None | OneOrMany[Tag] = None,
		code: int = 0
	) -> None:
		"""
		Insert/upsert tier docs into DB.
		"""

		if logger.verbose:
			if logger.verbose > 1:
				if body is None:
					logger.log(DEBUG, "T2 unit returned no body")
				else:
					logger.log(DEBUG, f'Saving T{self.tier} unit result', extra={"body": body})
			else:
				logger.log(VERBOSE, f'Saving T{self.tier} unit result')

		upd: JDict = {
			'$set': {'code': code},
			'$push': {'meta': meta}
		}

		if self.mtag:

			tag = merge_tags(self.mtag, tag) if tag else self.mtag # type: ignore
			activities = meta['activity']

			# T2 unit added a tag, make the distinction clear by adding a dedicated activity
			if [el for el in activities if 'tag' in el]:
				activities.append( # type: ignore[attr-defined]
					{'action': MetaActionCode.ADD_WORKER_TAG, 'tag': self.mtag}
				)
			else:
				activities[0]['action'] |= MetaActionCode.ADD_WORKER_TAG
				activities[0]['tag'] = self.mtag

		if tag:
			upd['$addToSet'] = {
				'tag': tag if isinstance(tag, (int, str)) else maybe_use_each(tag)
			}

		if body is not None:
			upd[payload_op]['body'] = body

		# Update document
		self.col.update_one(match, upd)


	def gen_meta(self,
		run_id: int,
		unit_trace_id: None | int,
		duration: int | float,
		action_code: MetaActionCode = MetaActionCode(0)
	) -> MetaRecord:

		d: MetaRecord = {
			'run': run_id,
			'ts': int(time()),
			'tier': self.tier,
			'code': None, # type: ignore[typeddict-item]
			'duration': duration,
			'activity': [{'action': action_code}],
			'traceid': {
				f't{self.tier}worker': self._trace_id,
				f't{self.tier}unit': unit_trace_id
			}
		}

		if self.job_sig:
			d['jobid'] = self.job_sig

		return d


	def sig_exit(self, signum: int, frame) -> None:
		""" Executed when SIGTERM/SIGINT is caught. Stops doc processing in run() """
		self._run = False


	def create_beacon(self) -> None:
		""" Creates a beacon document if it does not exist yet """

		args = {
			'class': self.__class__.__name__,
			'unit': self.unit_ids,
			'code': self.code_match,
			'base_log_flag': self.base_log_flag.__int__(),
			'doc_limit': self.doc_limit
		}

		self.beacon_id = build_unsafe_dict_id(args)

		# Create beacon doc if it does not exist yet
		self.col_beacon = self._ampel_db.get_collection('beacon')

		# Create specific beacon doc if it does not exist yet
		self.col_beacon.update_one(
			{'_id': self.beacon_id},
			{
				'$setOnInsert': {
					'class': args.pop('class'),
					'config': convert_dollars(args)
				},
				'$set': {'timestamp': int(time())}
			},
			upsert=True
		)


	def get_unit_instance(self,
		doc: T1Document | T2Document,
		logger: AmpelLogger
	) -> LogicalUnit:

		k = f'{doc["unit"]}_{doc["config"]}'

		# Check if T2 instance exists in this run
		if k not in self._instances:

			# Create channel (buffering) logger
			buf_hdlr = DefaultRecordBufferingHandler(level=logger.level)
			buf_hdlr._unit = doc['unit']
			buf_logger = AmpelLogger.get_logger(
				name = k,
				base_flag = (getattr(logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
				console = False,
				handlers = [buf_hdlr]
			)

			# Instantiate unit
			unit_instance = self._loader.new_logical_unit(
				model = UnitModel(unit=doc['unit'], config=doc['config']),
				logger = buf_logger
			)

			# Shortcut to avoid unit_instance.logger.handlers[?]
			setattr(unit_instance, '_buf_hdlr', buf_hdlr)
			self._instances[k] = unit_instance

		return self._instances[k]


def register_stats(tier: int) -> tuple[Histogram, Counter]:

	hist = AmpelMetricsRegistry.histogram(
		'latency',
		f'Delay between T{tier} doc creation and processing',
		subsystem=f't{tier}',
		unit='seconds',
		labelnames=('unit', ),
	)

	counter = AmpelMetricsRegistry.counter(
		'docs_processed',
		f'Number of T{tier} documents processed',
		subsystem=f't{tier}',
		labelnames=('unit', )
	)

	return hist, counter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/stock.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  20.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from ampel.types import Tag
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.time.QueryTimeModel import QueryTimeModel
from ampel.util.logicschema import reduce_to_set
from ampel.mongo.query.general import build_general_query


def build_stock_query(
	channel: None | int | str | dict | AllOf | AnyOf | OneOf = None,
	tag: None | dict[Literal['with', 'without'], Tag | dict | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag]] = None,
	time_created: None | QueryTimeModel = None,
	time_updated: None | QueryTimeModel = None
) -> dict[str, Any]:
	"""
	:param channel: string (one channel only) or a dict schema \
		(see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` \
		for syntax details). None (no criterium) means all channel are considered.

	:param with_tags: "tags" to be matched by query \
		(see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` syntax details). \

	:param without_tags: \
		"tags" not to be matched by query

	:param time_created: \
		match against the (channel dependant) transient document creation timestamp \
		(embedded in the transient journal)

	:param time_updated: \
		match against the (channel dependant) transient document modification timestamp \
		(embedded in the transient journal)

	:returns: query dict instance with adequate matching criteria
	:raises ValueError: build_general_query can raise ValueError \
		in case the provided dict schema structure is unsupported
	"""

	query = build_general_query(channel=channel, tag=tag)

	if time_created or time_updated:

		chans = reduce_to_set("any" if channel is None else channel)
		or_list = []

		for chan_name in chans:

			and_list = []

			if time_created:
				and_list.append({
					f'ts.{chan_name}.tied': time_created.dict()
				})

			if time_updated:
				and_list.append({
					f'ts.{chan_name}.upd': time_updated.dict()
				})

			or_list.append({'$and': and_list})

		query['$or'] = or_list

	return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/StartCommand.py
# License:             BSD-3-Clause
# Author:              jvs, vb
# Date:                13.03.2021
# Last Modified Date:  20.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import logging, signal, asyncio
from argparse import ArgumentParser
from typing import Any
from collections.abc import Sequence
from ampel.abstract.AbsCLIOperation import AbsCLIOperation
from ampel.core.AmpelController import AmpelController
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
#from ampel.log.AmpelLogger import AmpelLogger
#from ampel.log.LogFlag import LogFlag
from ampel.secret.AmpelVault import AmpelVault
from ampel.config.AmpelConfig import AmpelConfig


# Help parameter descriptions
h = {
	"config": "Path to an ampel config file (yaml/json)",
	"secrets": "Path to a YAML secrets store in sops format",
	"process": "Regex matching a process name. --process can be specified multiple times",
	"tier": "Run only processes of this tier",
	"exclude": "Specifically ecclude process(es) using this regex",
	"skip-validation": "Skip validation of unit models (safe to use if the ampel conf is trusted)",
	"verbose": "Increases verbosity"
}

class ScheduleCommand(AbsCLIOperation):
	"""
	TODO: implement properly
	"""

	@staticmethod
	def get_sub_ops() -> None | list[str]:
		return None

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		parser = AmpelArgumentParser()
		parser.set_help_descr(h)

		parser.req("config")
		parser.req("process", type=str, nargs="*", default=None)

		parser.opt("exclude", type=str, nargs="*", default=None)
		parser.opt("tier", action=MaybeIntAction, choices=(0, 1, 2, 3, "ops"), default=None)
		parser.opt("secrets", default=None)
		parser.opt("skip-validation", action='store_true')
		parser.opt("verbose", action="count", default=0)

		return parser


	# Mandatory implementation
	def run(self,
		args: dict[str, Any],
		unknown_args: Sequence[str],
		sub_op: None | str = None
	) -> None:

		self.args = args
		self.unknown_args = unknown_args
		# logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)
		config = AmpelConfig.load(args['config'])

		if (secrets_file := args['secrets']):
			from ampel.secret.DictSecretProvider import DictSecretProvider
			secrets = DictSecretProvider.load(secrets_file)

		self.el_capitan = AmpelController(
			config_arg = config,
			vault = AmpelVault(providers=[secrets]) if args['secrets'] else None,
			tier = args['tier'] if args['tier'] else None,
			match = args['process'],
			exclude = args['exclude'] if args['exclude'] else None,
			verbose = args['verbose']
		)

		if not self.el_capitan.controllers:
			print("No matching process found")
			return

		loop = asyncio.get_event_loop()
		task = loop.create_task(self.el_capitan.run())
		self.handle_signals(task, loop)
		loop.add_signal_handler(signal.SIGUSR1, self.reload_config)

		for result in loop.run_until_complete(task):
			if isinstance(result, asyncio.CancelledError):
				...
			elif isinstance(result, BaseException):
				raise result


	def handle_signals(self, task, loop, graceful=True):

		for s in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
			loop.remove_signal_handler(s)
			loop.add_signal_handler(
				s,
				lambda s=s, task=task, loop=loop, graceful=graceful: asyncio.create_task(
					self.shutdown(s, task, loop, graceful)
				),
			)


	async def shutdown(self, sig, task, loop, graceful=True):
		"""Stop root task on signal"""

		if graceful:
			logging.info(
				f"Received exit signal {sig.name}, shutting down gracefully (signal again to terminate immediately)..."
			)
			for controller in self.el_capitan.controllers:
				controller.stop()
			self.handle_signals(task, loop, False)
		else:
			logging.info(
				f"Received exit signal {sig.name}, terminating immediately..."
			)
			task.cancel()

		await task
		loop.stop()


	def reload_config(self) -> None:

		if self.args is None:
			return

		try:

			logging.info(f"Reloading config from {self.args['config']}")
			config = AmpelConfig.load(self.args['config'])

			secrets = None
			if self.args['secrets']:
				from ampel.secret.DictSecretProvider import DictSecretProvider
				secrets = DictSecretProvider.load(self.args['secrets'])

			if self.args['skip_validation']:
				import contextlib
				pyctx: Any = contextlib.nullcontext
			else:
				from ampel.core.UnitLoader import UnitLoader
				loader = UnitLoader(
					config,
					db=None,
					provenance=False,
					vault=AmpelVault(providers=[secrets]) if secrets else None
				)
				pyctx = loader.validate_unit_models

			with pyctx():
				groups = AmpelController.group_processes(
					AmpelController.get_processes(
						config, tier=self.args['tier'], match=self.args['match'],
					)
				)

		except Exception:
			logging.exception(f"Failed to load {self.args['config']}")
			return

		try:
			controllers = list(self.el_capitan.controllers)
			matches = []

			for group in groups:
				names = {pm.name for pm in group}
				for i, candidate in enumerate(controllers):
					if names.intersection([pm.name for pm in candidate.processes]):
						matches.append((candidate, group))
						del controllers[i]
						break
				else:
					raise RuntimeError(f"No match for process group {names}")
			assert len(matches) == len(self.el_capitan.controllers)

		except Exception:
			logging.exception("Failed to match process groups with current set")

		for controller, processes in matches:
			try:
				controller.update(config, loader.vault, processes)
				logging.info(
					f"Updated {controller.__class__.__name__} with processes: {[pm.name for pm in processes]} "
				)
			except Exception:
				logging.exception(
					f"Failed to update {controller.__class__.__name__} with processes: {[pm.name for pm in processes]}"
				)

		loop = asyncio.get_event_loop()
		task = loop.create_task(self.el_capitan.run())
		self.handle_signals(task, loop)
		loop.add_signal_handler(signal.SIGUSR1, self.reload_config)

		for result in loop.run_until_complete(task):
			if isinstance(result, asyncio.CancelledError):
				...
			elif isinstance(result, BaseException):
				raise result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3ProjectingStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.01.2020
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from collections.abc import Generator, Iterable, Sequence
from itertools import islice
from multiprocessing import JoinableQueue
from multiprocessing.pool import ThreadPool

from ampel.log import VERBOSE
from ampel.types import StockId, UBson
from ampel.t3.T3DocBuilder import AbsT3s
from ampel.view.T3Store import T3Store
from ampel.view.SnapView import SnapView
from ampel.struct.UnitResult import UnitResult
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.abstract.AbsT3Filter import AbsT3Filter
from ampel.abstract.AbsT3Projector import AbsT3Projector
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.t3.stage.T3ThreadedStager import T3ThreadedStager
from ampel.t3.stage.SimpleViewGenerator import SimpleViewGenerator
from ampel.t3.stage.NoViewGenerator import NoViewGenerator
from ampel.model.t3.T3ProjectionDirective import T3ProjectionDirective


class RunBlock:
	"""
	Used internally by T3UnitRunner
	"""
	filter: None | AbsT3Filter
	projector: None | AbsT3Projector
	units: list[AbsT3ReviewUnit]
	stock_ids: None | list[StockId]
	qdict: dict[type, list[JoinableQueue]]

	def __init__(self):
		self.filter = None
		self.projector = None
		self.units = []
		self.stock_ids = None
		self.len_view_types = 0
		self.qvals = None


class T3ProjectingStager(T3ThreadedStager):
	"""
	Default implementation handling the instructions defined
	by :class:`T3Directive.run <ampel.model.t3.T3Directive.T3Directive>`.
	"""

	#: Processing specification
	directives: Sequence[T3ProjectionDirective]

	#: Quick n Dirty: do not cast ampel buffers into views (thereby violating typing declarations)
	#: Added for the CLI operation 'ampel view show/save' as view casts are unnecessary in this case
	keep_buffers: bool = False


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.run_blocks: list[RunBlock] = []
		debug = self.logger.verbose > 1
		self.ViewGenerator: SimpleViewGenerator[SnapView] = NoViewGenerator if self.keep_buffers \
			else SimpleViewGenerator # type: ignore

		if debug:
			self.logger.debug("Setting up T3ProjectingStager")

		for directive in self.directives:

			rb = RunBlock()

			if directive.filter:

				if debug:
					self.logger.debug(f"Setting up filter {directive.filter.unit}")

				rb.filter = self.context.loader.new(
					model = directive.filter,
					unit_type = AbsT3Filter,
					logger = self.logger
				)

				if self.save_stock_ids:
					rb.stock_ids = []


			if directive.project:

				if debug:
					self.logger.debug(f"Setting up projector {directive.project.unit}")

				rb.projector = AuxUnitRegister.new_unit(
					model = directive.project,
					sub_type = AbsT3Projector,
					logger = self.logger
				)

			for exec_def in directive.execute:
				rb.units.append(
					self.get_unit(exec_def)
				)

			self.run_blocks.append(rb)


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		if len(self.run_blocks) == 1:
			if len(self.run_blocks[0].units) == 1:
				return self.proceed(
					self.run_blocks[0].units[0],
					self.ViewGenerator( # type: ignore[operator]
						self.run_blocks[0].units[0],
						self.projected_buffer_generator(self.run_blocks[0], gen),
						self.stock_updr
					),
					t3s
				)
			else:
				return self.proceed_threaded(
					self.run_blocks[0].units,
					self.projected_buffer_generator(self.run_blocks[0], gen),
					t3s
				)
		else:

			if self.chunk_size <= 0:
				raise ValueError("Chunking is required when multiple filter/projection blocks are defined")

			return self.multi_bla(gen, t3s)
			#for gen in tee(data, len(self.run_blocks)):
			# pass


	def projected_buffer_generator(self,
		run_block: RunBlock,
		gen: Generator[AmpelBuffer, None, None]
	) -> Generator[AmpelBuffer, None, None]:
	
		# Chunk input buffers (loaded from generator)
		buffers: Sequence[AmpelBuffer]

		try:
			while (buffers := list(islice(gen, self.chunk_size))):

				if run_block.filter:

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block filter")

					buffers = run_block.filter.filter(buffers)

					if self.save_stock_ids:
						run_block.stock_ids.extend([el['id'] for el in buffers])  # type: ignore[union-attr]

				if run_block.projector:

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block projection")

					buffers = run_block.projector.project(buffers)

				for el in buffers:
					yield el

		except RuntimeError as e:
			if "StopIteration" in str(e):
				return None
			raise e


	def multi_bla(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Generator[T3Document, None, None]:
		"""
		Handles multi run-blocks, that is multi projection/filters using the same input data
		"""

		all_units = []
		for rb in self.run_blocks:
			all_units += rb.units

		try:

			ts = time()
			with ThreadPool(processes=len(all_units)) as pool:

				# Create queues and generators for all instanciated t3 units
				queues, generators, async_results = self.create_threaded_generators(pool, all_units, t3s)

				# Add helper values to run_blocks
				for rb in self.run_blocks:

					# Number of view types for this this run block
					rb.len_view_types = len({t3_unit._View for t3_unit in rb.units})

					# Subset of queues for this run block
					rb.qvals = [q for u, q in queues.items() if u in rb.units]

					# Optimize by potentially grouping units associated with the same view type
					rb.qdict = {}
					for unit in rb.units:
						if unit.__class__._View not in rb.qdict:
							rb.qdict[unit.__class__._View] = []
						rb.qdict[unit.__class__._View].append(queues[unit])


				# Chunk input buffers (loaded from generator)
				while (x := list(islice(gen, self.chunk_size))):

					# Iterate over run blocks and add the projected buffers into the corresponding queues
					for rb in self.run_blocks:

						buffers: Iterable[AmpelBuffer] = x
						if rb.filter:

							if self.logger.verbose:
								self.logger.log(VERBOSE, "Applying run-block filter")

							buffers = rb.filter.filter(buffers)

							if self.save_stock_ids:
								rb.stock_ids.extend([el['id'] for el in buffers])  # type: ignore[union-attr]

						if rb.projector:

							if self.logger.verbose:
								self.logger.log(VERBOSE, "Applying run-block projection")

							buffers = rb.projector.project(buffers)

						self.put_views(buffers, rb.qdict)

					# Send sentinel all threaded generators
					for q in queues.values():
						q.put(None) # type: ignore[arg-type]

					for async_res, generator, t3_unit in zip(async_results, generators, all_units):

						# potential T3Record to be included in the T3Document
						if (t3_unit_result := async_res.get()):
							if (z := self.handle_t3_result(t3_unit, t3_unit_result, t3s, generator.stocks, ts)):
								yield z

				for i, rb in enumerate(self.run_blocks):
					self.flush(rb.units, extra={'directive': i})

		except Exception as e:
			self.flush(all_units)
			self.event_hdlr.handle_error(e, self.logger)


	def craft_t3_doc(self,
		t3_unit: AbsT3s,
		res: None | UBson | UnitResult,
		t3s: T3Store,
		ts: float,
		stocks: None | list[StockId] = None
	) -> T3Document:

		t3_doc = super().craft_t3_doc(t3_unit, res, t3s, ts, stocks)
		if self.save_stock_ids:
			rb = next(filter(lambda rb: t3_unit in rb.units, self.run_blocks))
			if rb.stock_ids:
				t3_doc['stock'] = rb.stock_ids
		return t3_doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/StockCompiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.05.2021
# Last Modified Date:  21.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ujson import encode
from typing import Any
from ampel.types import ChannelId, Tag, StockId
from ampel.content.StockDocument import StockDocument
from ampel.content.JournalRecord import JournalRecord
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsCompiler import AbsCompiler
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.base.AuxUnitRegister import AuxUnitRegister


class StockCompiler(AbsCompiler):
	"""
	Compiles updates to stock document arising at different ingester stages
	Note: mypy ignores are required as of May 2021 because mypy does not
	support generic typed dict (#3863) and higher kind typevars (#548)
	"""

	id_mapper: None | str


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.register: dict[StockId, dict[str, Any]] = {}
		self._id_mapper = AuxUnitRegister.get_aux_class(
			self.id_mapper, sub_type=AbsIdMapper
		) if self.id_mapper else None


	# Override
	def add(self, # type: ignore[override]
		stock: StockId,
		channel: ChannelId,
		journal: None | JournalRecord = None,
		tag: None | Tag | list[Tag] = None
	) -> None:

		if stock in self.register:
			d = self.register[stock]
			d['channel'].add(channel)
		else:
			d = self.register[stock] = {'stock': stock, 'channel': {channel}}
			if self._id_mapper:
				d['name'] = [self._id_mapper.to_ext_id(d['stock'])]

		if journal:
			# try to merge identical journal entries with each other
			# cannot use frozenset(items) because of potential nested dicts
			k = encode(journal)
			if 'journal' in d:
				if k in d['journal']:
					d['journal'][k][1].add(channel)
				else:
					d['journal'][k] = journal, {channel}
			else:
				d['journal'] = {k: (journal, {channel})}

		if tag:
			if 'tag' not in d:
				d['tag'] = set()
			if isinstance(tag, (str, int)):
				d['tag'].add(tag)
			else:
				d['tag'].update(tag)


	# Override
	def commit(self,
		ingester: AbsDocIngester[StockDocument],
		now: int | float,
		**kwargs
	) -> None:
		"""
		Note that we let the ingester handle 'tied' and 'upd' time values
		"""

		for k, v in self.register.items():

			d: StockDocument = {
				'stock': v['stock'],
				'channel': list(v['channel'])
			}

			if self._tag:
				if 'tag' in v:
					d['tag'] = self._tag + list(v['tag'])
				else:
					d['tag'] = self._tag
			elif 'tag' in v:
				d['tag'] = v['tag']

			if 'name' in v:
				d['name'] = v['name']

			if kwargs.get('body'):
				d['body'] = kwargs['body']

			if self.origin:
				d['origin'] = self.origin

			if 'journal' in v:
				d['journal'] = [
					{
						'ts': now, 'run': self.run_id,
						'tier': self.tier, 'channel': list(chans)
					} | entry
					for entry, chans in v['journal'].values()
				]

			ingester.ingest(d)

		self.register.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3ChannelStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                21.06.2020
# Last Modified Date:  09.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator, Sequence
from ampel.types import ChannelId
from ampel.view.T3Store import T3Store
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.model.UnitModel import UnitModel
from ampel.model.t3.T3ProjectionDirective import T3ProjectionDirective
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.t3.stage.T3ProjectingStager import T3ProjectingStager


class T3ChannelStager(AbsT3Stager):
	"""
	Stager class that configures an underlying T3ProjectingStager instance
	to filter and project AmpelBuffer instances with respect to a single channel.
	Essentially a shortcut class since the same functionality
	can be obtained using a properly configured T3ProjectingStager instance.
	T3ProjectingStager is not inherited because of the differences in model definitions.
	"""

	channel: ChannelId
	execute: Sequence[UnitModel]


	def __init__(self, **kwargs) -> None:
		"""
		:param channel: channel name/id
		:param execute: sequence of T3 units to be executed
		"""

		super().__init__(
			**{k: v for k, v in kwargs.items() if k in self._annots}
		)

		self._stager = T3ProjectingStager(
			directives=[
				T3ProjectionDirective(
					filter = UnitModel(
						unit="T3AmpelBufferFilter",
						config={"channel": kwargs['channel']}
					),
					project = UnitModel(
						unit = "T3ChannelProjector",
						config = {"channel": kwargs['channel']}
					),
					execute = kwargs.pop('execute')
				)
			],
			**kwargs
		)


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:
		return self._stager.stage(gen, t3s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/aux/filter/FlatDictArrayFilter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import MutableMapping
from collections.abc import Mapping, Sequence
from ampel.util.mappings import flatten_dict, unflatten_dict
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.aux.filter.AbsLogicOperatorFilter import AbsLogicOperatorFilter
from ampel.aux.filter.SimpleDictArrayFilter import SimpleDictArrayFilter


class FlatDictArrayFilter(AbsLogicOperatorFilter[MutableMapping]):
	"""
	In []: f = FlatDictArrayFilter(filters={'attribute': 'a.b', 'operator': '>', 'value': 2})
	In []: f.apply([{'a': {'b': 1}}, {'a': {'b':4}}])
	Out[]: [{'a': {'b': 4}}]
	"""

	@staticmethod
	def _apply_filter(dicts: Sequence[Mapping], f: FilterCriterion) -> list[MutableMapping]:
		return [
			unflatten_dict(ell)
			for ell in SimpleDictArrayFilter._apply_filter([flatten_dict(el) for el in dicts], f)
		]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ampel',
 'ampel.abstract',
 'ampel.aux',
 'ampel.aux.filter',
 'ampel.cli',
 'ampel.config',
 'ampel.config.builder',
 'ampel.config.collector',
 'ampel.core',
 'ampel.demo',
 'ampel.dev',
 'ampel.ingest',
 'ampel.log',
 'ampel.log.handlers',
 'ampel.metrics',
 'ampel.model',
 'ampel.model.aux',
 'ampel.model.builder',
 'ampel.model.ingest',
 'ampel.model.job',
 'ampel.model.purge',
 'ampel.model.t3',
 'ampel.model.time',
 'ampel.mongo',
 'ampel.mongo.model',
 'ampel.mongo.purge',
 'ampel.mongo.query',
 'ampel.mongo.query.var',
 'ampel.mongo.update',
 'ampel.mongo.update.var',
 'ampel.mongo.view',
 'ampel.ops',
 'ampel.run',
 'ampel.secret',
 'ampel.t1',
 'ampel.t2',
 'ampel.t3',
 'ampel.t3.include.session',
 'ampel.t3.stage',
 'ampel.t3.stage.filter',
 'ampel.t3.stage.project',
 'ampel.t3.supply',
 'ampel.t3.supply.complement',
 'ampel.t3.supply.load',
 'ampel.t3.supply.select',
 'ampel.t3.unit',
 'ampel.template',
 'ampel.test',
 'ampel.util',
 'ampel.vendor.aiopipe']

package_data = \
{'': ['*'], 'ampel.test': ['test-data/*']}

install_requires = \
['ampel-interface>=0.8.3-beta.11,<0.9.0',
 'appdirs>=1.4.4,<2.0.0',
 'prometheus-client>=0.10,<0.12',
 'psutil>=5.8.0,<6.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pymongo>=4.0,<5.0',
 'schedule>=1.0.0,<2.0.0',
 'sjcl>=0.2.1,<0.3.0',
 'xxhash>=3.0.0,<4.0.0',
 'yq>=3.0.0,<4.0.0']

extras_require = \
{'docs': ['Sphinx>=5.1.1,<5.2.0',
          'sphinx-press-theme>=0.5.1,<0.9.0',
          'sphinx-autodoc-typehints>=1.11.1,<2.0.0',
          'tomlkit>=0.7.0,<0.8.0'],
 'server': ['fastapi>=0.63,<0.69', 'uvicorn[standard]>=0.13.3,<0.16.0'],
 'slack': ['slack-sdk>=3.18.1,<4.0.0']}

entry_points = \
{'cli': ['buffer_Match_and_view_or_save_ampel_buffers = '
         'ampel.cli.BufferCommand',
         'config_Build_or_update_config._Fetch_or_append_config_elements = '
         'ampel.cli.ConfigCommand',
         'db_Initialize,_dump,_delete_specific_databases_or_collections = '
         'ampel.cli.DBCommand',
         'job_Run_schema_file(s) = ampel.cli.JobCommand',
         'log_Select,_format_and_either_view_(tail_mode_available)_or_save_logs '
         '= ampel.cli.LogCommand',
         'process_Run_single_task = ampel.cli.ProcessCommand',
         'run_Run_selected_process(es)_from_config = ampel.cli.RunCommand',
         't2_Match_and_either_reset_or_view_raw_t2_documents = '
         'ampel.cli.T2Command',
         'view_Select,_load_and_save_fresh_"ampel_views" = '
         'ampel.cli.ViewCommand'],
 'console_scripts': ['ampel = ampel.cli.main:main',
                     'ampel-controller = '
                     'ampel.core.AmpelController:AmpelController.main',
                     'ampel-db = ampel.db.AmpelDB:main']}

setup_kwargs = {
    'name': 'ampel-core',
    'version': '0.8.3b15',
    'description': 'Alice in Modular Provenance-Enabled Land',
    'long_description': '<img align="left" src="https://desycloud.desy.de/index.php/s/99Jkcyzn92rRpHF/preview" width="150" height="150"/>\n<br>\n\n# AMPEL-core\n<br><br>\n\n# Introduction\n\nAMPEL is a _modular_ and _scalable_ platform with explicit _provenance_ tracking, suited for systematically processing large - possibly complex and heterogeneous - datasets in real-time or not. This includes selecting, analyzing, updating, combining, enriching and reacting to data.\n\nThe framework requires analysis and reaction logic to be broken down in adequate indepentent units.\nAMPEL is general enough to be applicable in various fields,\nit was originaly developped to solve challenges in the context of experimental astrophysics.\n\nAMPEL is written in Python 3.10 and its codebase is fully typed.\n \n\n# Architecture\n\n## Tiers\nAMPEL is made of four execution layers (tiers) that replace a traditional pipeline architecture.\n<img align="left" src="https://desycloud.desy.de/index.php/s/fz2mnsH4MGEKwfD/preview"/>\n\nThe tiers are independently scheduled and the information exchange between tiers occurs via a dedicated database.\nThe execution layer architecture along with the database structure allows for simple parallelization.\n\n## Units\nEach tier is modular and executes so-called "units".\n\n<p align="center">\n<img src="https://desycloud.desy.de/index.php/s/P76f9qSWJse8oT7/preview" width=50%/>\n</p>\n\nAmpel _base units_ have standardized inputs and ouputs, enforced through abstract classes which units inherit.\n\n## Processes\n\nEvery change in AMPEL is triggered by a _process_.\nA process executes, at a given date and time, a _processor unit_ that itself runs one or multiple _base units_ with specific configurations.\nInformation about process executions are registred into the database.\nThe majority of processes are associated with a specific tier but general processes are possible.\n\nA working AMPEL system will spawn multiple processes, posssibly concurently, accross the four AMPEL tiers.\nThis will result in the ingestion and analysis of data and the triggering of automated reactions when given data states are detected.\n\n## Channels\n\n_Channels_ are convenient for multi-user or multi-prupose AMPEL systems.\nThey allow to define and enforce access rights and to privatize processes,\nmeaning that the output generated by the processes will be only accessible to processes\nbelonging to the same channel.\n\nInternally, _channels_ are just tags in database documents and ampel configuration files.  \nFrom a user perspective, a channel can be seen as a collection of private processes.\n\n<p align="center">\n<img src="https://desycloud.desy.de/index.php/s/YMiGJ2zckgEr54n/preview" width=50%/>\n<br/>\nProcesses associated with a given channel\n</p>\n\nNote that within AMPEL, different _channels_ requiring the same computation\nwill not result in the required computation being performed twice.\n\n\n# Repositories\n\nThe AMPEL code is partitioned in different repositories.  \nThe only mandatory repository in this list is _ampel-interface_\n\nPublic abstract class definitions:  \nhttps://github.com/AmpelProject/Ampel-interface\n\nSpecialized classes for Tier 0, capable of handling _alerts_:  \nhttps://github.com/AmpelProject/Ampel-alerts\n\nAn add-on that introduces two generic classes of datapoints:  \nhttps://github.com/AmpelProject/Ampel-photometry\n\nExample of an instrument specific implementation:  \nhttps://github.com/AmpelProject/Ampel-ztf\n\nNumerous _base units_, the majority being specific to astronomy:  \nhttps://github.com/AmpelProject/Ampel-contrib-HU/\n\n\n# Database\n\nMongoDB is used to store data.\nThe collections have been designed and indexed for fast insertion and query.\nUsers do not interact with the database directly.\nInformation exchange is instead regulated through (python) abstract base classes from which units are constructed.\nA specific set of internal classes handle database input and output.\n\n\n# Containers\n\nAll AMPEL software, can be combined into one container that defines an instance.\nThese containers can be used both to process real-time data as well as to reprocess archived data.\nThe containers themselves should be archived as well.\n\n<!--\nAstronomers have during the past century continuously refined tools for\nanalyzing individual astronomical transients. Simultaneously, progress in instrument and CCD\nmanufacturing as well as new data processing capabilities have led to a new generation of transient\nsurveys that can repeatedly scan large volumes of the Universe. With thousands of potential candidates\navailable, scientists are faced with a new kind of questions: Which transient should I focus on?\nWhat were those things that I dit not look at? Can I have them all?\n\nAmpel is a software framework meant to assist in answering such questions.\nIn short, Ampel assists in the the transition from studies of individual objects\n(based on more or less random selection) to systematically selected samples.\nOur design goals are to find a system where past experience (i.e. existing algorithms and code) can consistently be applied to large samples, and with built-in tools for controlling sample selection.\n-->\n\n# Installing Ampel\n\nThe latest release of `ampel-core` can be installed from PyPI with `pip`, e.g.:\n\n```\npip install ampel-core\n```\n\nOther projects like `ampel-alerts`, `ampel-ztf`, etc. can also be installed with `pip`.\n\n# Development\n\n`ampel-core` uses [poetry](http://poetry.eustace.io/) for dependency management and packaging. To work with an editable install it\'s recommended that you set up `poetry` and install `ampel-core` in a virtual environment by doing\n\n```console\npoetry install\n```\n\nAlternatively, may also use a `setuptools`-style editable install from `setup.py`:\n\n```console\npip install -e .\n```\n\nNote that `setup.py` and `requirements.txt` are auto-generated; any changes you commit will be overwritten the next time `pyproject.toml` is updated.\n',
    'author': 'Valery Brinnel',
    'author_email': 'None',
    'maintainer': 'Jakob van Santen',
    'maintainer_email': 'jakob.van.santen@desy.de',
    'url': 'https://ampelproject.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



import pickle
from pathlib import Path

import pytest

from ampel.content.StockDocument import StockDocument
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.log.AmpelLogger import AmpelLogger, DEBUG
from ampel.t3.stage.project.T3ChannelProjector import T3ChannelProjector


@pytest.fixture
def stock_doc() -> StockDocument:
    with open(Path(__file__).parent / "test-data" / "ZTF20abxvcrk.pkl", "rb") as f:
        doc = pickle.load(f)
    # convert from v0.7 to v0.8
    doc["stock"] = doc["_id"]
    doc["ts"] = {
        channel: {
            "tied": doc["created"][channel],
            "upd": doc["modified"][channel],
        } for channel in doc["created"].keys() if channel != "any"
    }
    del doc["created"]
    doc["updated"] = doc.pop("modified")["any"]
    return doc


@pytest.fixture
def logger():
    return AmpelLogger.get_logger(console={"level": DEBUG})


def strip_channel(jentries):
    return [{k: v for k, v in jentry.items() if k != "channel"} for jentry in jentries]


def test_single_channel(stock_doc: StockDocument, logger):
    target = "TDE_RANKING"
    before = AmpelBuffer(id=stock_doc["stock"], stock=stock_doc)
    input_stock_doc = before["stock"]
    assert input_stock_doc is not None
    all_channels = set(input_stock_doc["channel"])
    assert target in all_channels, "target is in test case channel set"
    assert len(all_channels) > 1, "test case contains channels not in target"

    proj = T3ChannelProjector(channel=target, logger=logger)
    after = proj.project([before])[0]
    output_stock_doc = after["stock"]
    assert output_stock_doc is not None
    for field in "tag", "name", "channel":
        assert not isinstance(output_stock_doc[field], str), f"stock.{field} must be a set" # type: ignore
    assert output_stock_doc["channel"] == [target]
    assert output_stock_doc["ts"] == {target: input_stock_doc["ts"][target]}

    before_no_channel = strip_channel(input_stock_doc["journal"])
    after_no_channel = strip_channel(output_stock_doc["journal"])

    # all entries have only the selected channel
    for jentry in output_stock_doc["journal"]:
        if "channel" in jentry:
            assert jentry["channel"] == target

    for jentry, stripped in zip(input_stock_doc["journal"], before_no_channel):
        if "channel" in jentry:
            try:
                after_no_channel.index(
                    stripped
                ), "entry with target channel was preserved"
            except ValueError:
                if isinstance(jentry["channel"], int):
                    if target == jentry["channel"]:
                        raise
                elif target in jentry["channel"]:
                    raise


@pytest.mark.parametrize("logic_op", ["all_of", "any_of", "one_of"])
def test_multi_channel(stock_doc, logger, logic_op):
    target = {"RCF_2020B", "ZTF_SLSN"}
    before = AmpelBuffer(id=stock_doc["_id"], stock=stock_doc)
    all_channels = set(before["stock"]["channel"])
    assert target.intersection(all_channels), "target is in test case channel set"
    assert len(all_channels) > len(target), "test case contains channels not in target"

    proj = T3ChannelProjector(channel={logic_op: list(target)}, logger=logger)
    after = proj.project([before])[0]
    for field in "tag", "name", "channel":
        assert not isinstance(after["stock"][field], str), f"stock.{field} must be a set"
    assert set(after["stock"]["channel"]) == target
    assert after["stock"]["updated"] == before["stock"]["updated"]
    assert after["stock"]["ts"] == {k: before["stock"]["ts"][k] for k in target}

    before_no_channel = strip_channel(before["stock"]["journal"])
    after_no_channel = strip_channel(after["stock"]["journal"])

    # all entries have only the selected channel
    for jentry in after["stock"]["journal"]:
        if (channel := jentry.get("channel")) is None:
            continue
        if isinstance(channel, (int, str)):
            assert channel in target
        else:
            assert len(channel) > 1
            assert set(channel).issubset(target)

    for jentry, stripped in zip(before["stock"]["journal"], before_no_channel):
        if (channel := jentry.get("channel")) is None:
            continue
        try:
            after_no_channel.index(
                stripped
            ), "entry with channel in target was preserved"
        except ValueError:
            if (isinstance(channel, (int, str)) and channel in target) or set(
                channel
            ).issubset(target):
                raise
